from allencv.modules.encoder_base import _EncoderBase

from allennlp.common import Registrable


class Im2VecEncoder(_EncoderBase, Registrable):

    def get_input_channels(self) -> int:
        raise NotImplementedError

    def get_output_dim(self) -> int:
        raise NotImplementedError



from typing import Dict, List, Tuple
import logging
from overrides import overrides

import torch
import torch.nn.functional as F
from torchvision.models.detection.rpn import AnchorGenerator, concat_box_prediction_layers
from torchvision.models.detection.image_list import ImageList
from torchvision.models.detection import _utils as det_utils
from torchvision.ops import boxes as box_ops
from torchvision.models.detection import fasterrcnn_resnet50_fpn

from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator
from allennlp.training.metrics import Average

from allencv.nn.common import RPNHead
from allencv.modules.image_encoders import ImageEncoder, ResnetEncoder, FPN
from allencv.models.object_detection import utils as object_utils

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@Model.register("rpn")
class RPN(Model):
    """
    A region proposal network identifies regions of an input image that correspond
    to objects of various classes. It is typically used as the first stage in a
    Faster R-CNN.

    Parameters
    ----------
    backbone: ``ImageEncoder``
        Encoder network that provides features from the input image at various scales.
    positive_fraction: ``float``
        What fraction of each batch in the loss computation should be positive
        examples (foreground).
    match_thresh_low: ``float``
        Any anchor boxes that overlap with target boxes with an IOU threshold less than
        this number will be considered background.
    match_thresh_high: ``float``
        Any anchor boxes that overlap with target boxes with an IOU threshold greater than
        this number will be considered foreground.
    anchor_sizes: ``List[int]``
        Size of anchor boxes at various levels.
    anchor_aspect_ratios: ``List[float]``
        Which aspect ratios to include as anchors for each anchor location.
    anchor_strides: ``List[int]``
        How far apart each anchor box center is from its neighbors.
    batch_size_per_image: ``int``
        The number of examples to include in the loss function for each image in each batch.
    pre_nms_top_n: ``int``
        The number of proposals to keep before non-maximum suppression is applied.
    post_nms_top_n: ``int``
        The number of proposals after non-maximum suppression is applied.
    nms_thresh: ``float``
        Proposal boxes that overlap by more than this amount will be suppressed.
    min_size: ``int``
        The minimum size of a proposal bounding box.
    fpn_post_nms_top_n: ``int``
        How many proposals to keep after non-max supprossion is applied.
    fpn_post_nms_per_batch: ``bool``
        Whether the `fpn_post_nms_top_n` corresponds to each batch or each image.
    allow_low_quality_matches: ``bool``
        Produce additional matches for predictions that have only low-quality matches.
    straddle_thresh: ``int``
        How many pixels of an anchor box can be outside the image bounds before it is no
        longer considered valid.
    """

    def __init__(self,
                 backbone: ImageEncoder,
                 positive_fraction: float = 0.5,
                 match_thresh_low: float = 0.3,
                 match_thresh_high: float = 0.7,
                 anchor_sizes: List[int] = (128, 256, 512),
                 anchor_aspect_ratios: List[float] = (0.5, 1.0, 2.0),
                 anchor_strides: List[int] = (8, 16, 32),
                 batch_size_per_image: int = 256,
                 pre_nms_top_n: int = 6000,
                 post_nms_top_n: int = 300,
                 nms_thresh: int = 0.7,
                 min_size: int = 0,
                 fpn_post_nms_top_n: int = 1000,
                 fpn_post_nms_per_batch: int = True,
                 allow_low_quality_matches: bool = True,
                 straddle_thresh: int = 0,
                 initializer: InitializerApplicator = InitializerApplicator()) -> None:
        super(RPN, self).__init__(None)
        self._rpn_head = RPNHead(256, 3)
        self.min_size = min_size
        self.pre_nms_top_n = pre_nms_top_n
        self.nms_thresh = nms_thresh
        self.post_nms_top_n = post_nms_top_n

        # the BoxCoder just converts the relative regression offsets into absolute
        # coordinates
        self.box_coder = det_utils.BoxCoder(weights=(1., 1., 1., 1.))

        # sampler is responsible for selecting a subset of anchor boxes for computing the loss
        # this makes sure each batch has reasonable balance of foreground/background labels
        # it selects `batch_size_per_image` total boxes
        self.sampler = det_utils.BalancedPositiveNegativeSampler(batch_size_per_image, positive_fraction)

        # matcher decides if an anchor box is a foreground or background based on how much
        # it overlaps with the nearest target box
        self.proposal_matcher = det_utils.Matcher(
                match_thresh_high,
                match_thresh_low,
                allow_low_quality_matches=allow_low_quality_matches)

        self.backbone = backbone
        self.anchor_generator = AnchorGenerator(anchor_sizes, anchor_aspect_ratios)
        self.num_anchors = self.anchor_generator.num_anchors_per_location()[0]

        # initializer(self)

        # TODO: use initializer?
        # for l in [self.conv, self.cls_logits, self.bbox_pred]:
        #     torch.nn.init.normal_(l.weight, std=0.01)
        #     torch.nn.init.constant_(l.bias, 0)
        self._loss_meters = {'rpn_cls_loss': Average(), 'rpn_reg_loss': Average()}

    def assign_targets_to_anchors(self,
                                  anchors: List[torch.Tensor],
                                  targets: List[torch.Tensor]) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        labels = []
        matched_gt_boxes = []
        for anchors_per_image, gt_boxes in zip(anchors, targets):
            match_quality_matrix = box_ops.box_iou(gt_boxes, anchors_per_image)
            matched_idxs = self.proposal_matcher(match_quality_matrix)
            # get the targets corresponding GT for each proposal
            # NB: need to clamp the indices because we can have a single
            # GT in the image, and matched_idxs can be -2, which goes
            # out of bounds
            matched_gt_boxes_per_image = gt_boxes[matched_idxs.clamp(min=0)]

            labels_per_image = matched_idxs >= 0
            labels_per_image = labels_per_image.to(dtype=torch.float32)

            # Background (negative examples)
            bg_indices = matched_idxs == self.proposal_matcher.BELOW_LOW_THRESHOLD
            labels_per_image[bg_indices] = 0

            # discard indices that are between thresholds
            inds_to_discard = matched_idxs == self.proposal_matcher.BETWEEN_THRESHOLDS
            labels_per_image[inds_to_discard] = -1

            labels.append(labels_per_image)
            matched_gt_boxes.append(matched_gt_boxes_per_image)
        return labels, matched_gt_boxes

    def _get_top_n_idx(self, objectness, num_anchors_per_level):
        result = []
        offset = 0
        for ob in objectness.split(num_anchors_per_level, 1):
            num_anchors = ob.shape[1]
            pre_nms_top_n = min(self.pre_nms_top_n, num_anchors)
            _, top_n_idx = ob.topk(pre_nms_top_n, dim=1)
            result.append(top_n_idx + offset)
            offset += num_anchors
        return torch.cat(result, dim=1)

    def forward(self,
                image: torch.Tensor,  # (batch_size, c, h, w)
                image_sizes: torch.Tensor,  # (batch_size, 2)
                boxes: torch.Tensor = None,  # (batch_size, max_boxes_in_batch, 4)
                box_classes: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        # pylint: disable=arguments-differ
        im_sizes = [(x[1].item(), x[0].item()) for x in image_sizes]
        image_list = ImageList(image, im_sizes)
        features = self.backbone.forward(image)
        objectness, rpn_box_regression = self._rpn_head(features)
        anchors: List[torch.Tensor] = self.anchor_generator(image_list, features)
        num_anchors_per_level = [o[0].numel() for o in objectness]
        objectness, rpn_box_regression = \
            concat_box_prediction_layers(objectness, rpn_box_regression)

        out = {'features': features,
               'objectness': objectness,
               'rpn_box_regression': rpn_box_regression,
               'anchors': anchors,
               'sizes': image_sizes,
               'num_anchors_per_level': num_anchors_per_level}
        if boxes is not None:
            labels, matched_gt_boxes = self.assign_targets_to_anchors(
                    anchors, object_utils.padded_tensor_to_tensor_list(boxes))
            regression_targets = self.box_coder.encode(matched_gt_boxes, anchors)

            sampled_pos_inds, sampled_neg_inds = self.sampler(labels)
            sampled_pos_inds = torch.nonzero(torch.cat(sampled_pos_inds, dim=0)).squeeze(1)
            sampled_neg_inds = torch.nonzero(torch.cat(sampled_neg_inds, dim=0)).squeeze(1)

            sampled_inds = torch.cat([sampled_pos_inds, sampled_neg_inds], dim=0)

            objectness = objectness.flatten()

            labels = torch.cat(labels, dim=0)
            regression_targets = torch.cat(regression_targets, dim=0)

            loss_rpn_box_reg = F.l1_loss(
                    rpn_box_regression[sampled_pos_inds],
                    regression_targets[sampled_pos_inds],
                    reduction="sum",
            ) / (sampled_inds.numel())

            loss_objectness = F.binary_cross_entropy_with_logits(
                    objectness[sampled_inds], labels[sampled_inds]
            )
            self._loss_meters['rpn_cls_loss'](loss_objectness.item())
            self._loss_meters['rpn_reg_loss'](loss_rpn_box_reg.item())
            out["loss_objectness"] = loss_objectness
            out["loss_rpn_box_reg"] = loss_rpn_box_reg
            out["loss"] = loss_objectness + 10*loss_rpn_box_reg
        return out

    def filter_proposals(self, proposals, objectness, image_shapes, num_anchors_per_level):
        num_images = proposals.shape[0]
        device = proposals.device
        # do not backprop throught objectness
        objectness = objectness.detach()
        objectness = objectness.reshape(num_images, -1)

        levels = [
                torch.full((n,), idx, dtype=torch.int64, device=device)
                for idx, n in enumerate(num_anchors_per_level)
        ]
        levels = torch.cat(levels, 0)
        levels = levels.reshape(1, -1).expand_as(objectness)

        # select top_n boxes independently per level before applying nms
        top_n_idx = self._get_top_n_idx(objectness, num_anchors_per_level)
        batch_idx = torch.arange(num_images, device=device)[:, None]
        objectness = objectness[batch_idx, top_n_idx]
        levels = levels[batch_idx, top_n_idx]
        proposals = proposals[batch_idx, top_n_idx]

        final_boxes = []
        final_scores = []
        for boxes, scores, lvl, img_shape in zip(proposals, objectness, levels, image_shapes):
            boxes = box_ops.clip_boxes_to_image(boxes, img_shape)
            keep = box_ops.remove_small_boxes(boxes, self.min_size)
            boxes, scores, lvl = boxes[keep], scores[keep], lvl[keep]
            # non-maximum suppression, independently done per level
            keep = box_ops.batched_nms(boxes, scores, lvl, self.nms_thresh)
            # keep only topk scoring predictions
            keep = keep[:self.post_nms_top_n]
            boxes, scores = boxes[keep], scores[keep]
            final_boxes.append(boxes)
            final_scores.append(scores)
        return final_boxes, final_scores

    @overrides
    def decode(self, output_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        # you get a list of proposal boxes for each image in the batch
        proposals = self.box_coder.decode(
                output_dict['rpn_box_regression'], output_dict['anchors'])
        num_images = len(output_dict['anchors'])
        proposals = proposals.view(num_images, -1, 4)
        boxes, scores = self.filter_proposals(proposals,
                                              output_dict['objectness'],
                                              output_dict['sizes'],
                                              output_dict['num_anchors_per_level'])
        output_dict['proposals'] = boxes
        output_dict['proposal_scores'] = scores
        return output_dict

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = {k: v.get_metric(reset) for k, v in self._loss_meters.items()}
        return metrics


# @Model.register("pretrained")
# class PretrainedRPN(RPN):
#
#     @classmethod
#     def from_params(cls, params: Params):
#         model = load_archive(params.pop("archive_file")).model
#         requires_grad = params.pop("requires_grad")
#         for p in model.parameters():
#             p.requires_grad = requires_grad
#         return model
#
#
@Model.register("detectron_rpn")
class PretrainedDetectronRPN(RPN):

    def __init__(self,
                 anchor_sizes: List[int] = (128, 256, 512),
                 anchor_aspect_ratios: List[float] = (0.5, 1.0, 2.0),
                 anchor_strides: List[int] = (8, 16, 32),
                 batch_size_per_image: int = 256):
        backbone = ResnetEncoder('resnet50')
        fpn_backbone = FPN(backbone, 256)
        super(PretrainedDetectronRPN, self).__init__(fpn_backbone,
                                                     anchor_strides=anchor_strides,
                                                     anchor_aspect_ratios=anchor_aspect_ratios,
                                                     anchor_sizes=anchor_sizes,
                                                     batch_size_per_image=batch_size_per_image)
        frcnn = fasterrcnn_resnet50_fpn(pretrained=True, pretrained_backbone=True)
        backbone.load_state_dict(frcnn.backbone.body.state_dict())
        self._rpn_head.load_state_dict(frcnn.rpn.head.state_dict())

        # pylint: disable = protected-access
        fpn_backbone._convert_layers.load_state_dict(frcnn.backbone.fpn.inner_blocks.state_dict())
        fpn_backbone._combine_layers.load_state_dict(frcnn.backbone.fpn.layer_blocks.state_dict())



import torch

from allennlp.modules import FeedForward

from allencv.modules.im2vec_encoders.im2vec_encoder import Im2VecEncoder


@Im2VecEncoder.register("flatten")
class FlattenEncoder(Im2VecEncoder):
    """
    An ``Im2VecEncoder`` that simply flattens the input image and optionally passes it
    through a feedforward network.

    Parameters
    ----------
    input_channels: ``int``
        Number of channels in the input image.
    input_height: ``int``
        Height of input image.
    input_width: ``int``
        Width of input image.
    feedforward: ``FeedForward``
        Optional feedforward, dense network to pass the flattened vector through.
    """

    def __init__(self,
                 input_channels: int,
                 input_height: int,
                 input_width: int,
                 feedforward: FeedForward = None) -> None:
        super(FlattenEncoder, self).__init__()
        self.input_channels = input_channels
        self.input_height = input_height
        self.input_width = input_width
        self.feedforward = feedforward

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        output = inputs.view(inputs.shape[0], -1)
        if self.feedforward is not None:
            output = self.feedforward(output)
        return output

    def get_input_channels(self) -> int:
        return self.input_channels

    def get_output_dim(self) -> int:
        return self.input_height * self.input_width * self.input_channels \
            if self.feedforward is None else self.feedforward.get_output_dim()



from typing import Type

from allencv.data.transforms.image_transform import ImageTransform

from allennlp.common import Params

import albumentations as aug
import albumentations.augmentations.functional as af
from albumentations.core.transforms_interface import BasicTransform, ImageOnlyTransform


class _ImageTransformWrapper(object):

    def __init__(self, transform: Type[BasicTransform]):
        self._transform_class = transform

    def from_params(self, params: Params):
        transform = self._transform_class(**params.as_dict())
        return ImageTransform(transform)


class BGRNormalize(ImageOnlyTransform):

    def __init__(self,
                 mean=[102.9801, 115.9465, 122.7717],
                 std=[1., 1., 1.],
                 max_pixel_value=255.0,
                 always_apply=False, p=1.0):
        super(BGRNormalize, self).__init__(always_apply, p)
        self.mean = mean
        self.std = std
        self.max_pixel_value = max_pixel_value

    def apply(self, image, **params):
        image = image[:, :, [2, 1, 0]] * 255.
        return af.normalize(image, self.mean, self.std, self.max_pixel_value)

ImageTransform.register("resize")(_ImageTransformWrapper(aug.Resize))
ImageTransform.register("rotate")(_ImageTransformWrapper(aug.Rotate))
ImageTransform.register("gaussian_blur")(_ImageTransformWrapper(aug.GaussianBlur))
ImageTransform.register("elastic_transform")(_ImageTransformWrapper(aug.ElasticTransform))
ImageTransform.register("grid_distortion")(_ImageTransformWrapper(aug.GridDistortion))
ImageTransform.register("random_bbox_safe_crop")(_ImageTransformWrapper(aug.RandomSizedBBoxSafeCrop))
ImageTransform.register("random_brightness_contrast")(_ImageTransformWrapper(aug.RandomBrightnessContrast))
ImageTransform.register("bgr_normalize")(_ImageTransformWrapper(BGRNormalize))
ImageTransform.register("smallest_max_size")(_ImageTransformWrapper(aug.SmallestMaxSize))
ImageTransform.register("flip")(_ImageTransformWrapper(aug.Flip))
ImageTransform.register("horizontal_flip")(_ImageTransformWrapper(aug.HorizontalFlip))
ImageTransform.register("channel_shuffle")(_ImageTransformWrapper(aug.ChannelShuffle))
ImageTransform.register("normalize")(_ImageTransformWrapper(aug.Normalize))


from overrides import overrides
import numpy as np
from typing import Dict, Sequence

import torch

from allennlp.data.fields import Field


class ImageField(Field[np.array]):
    """
    An ``ImageField`` stores an image as a ``np.ndarray`` which must have exactly three
    dimensions.

    Parameters
    ----------
    image: ``np.ndarray``
    channels_first: ``bool``, optional (default=True)
        Whether the first or last dimension is the channel dimension.
    """
    def __init__(self,
                 image: np.ndarray,
                 channels_first: bool = True,
                 padding_value: int = 0) -> None:
        if len(image.shape) == 2:
            image = image[np.newaxis, :, :]
            channels_first = True
        self.image = image
        if not channels_first:
            h, w, c = self.image.shape
            self._layout = 'hwc'
        else:
            c, h, w = self.image.shape
            self._layout = 'chw'
        self._channels = c
        self._height = h
        self._width = w
        assert len(image.shape) == 3
        self.padding_value = padding_value
        self.channels_first = channels_first

    @overrides
    def get_padding_lengths(self) -> Dict[str, int]:
        return {
            'channels': self._channels,
            'height': self._height,
            'width': self._width
        }

    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        np_img = self.image
        img = torch.from_numpy(np_img).float()
        if self._layout == 'hwc':
            pad_img = torch.zeros(padding_lengths['height'],
                                  padding_lengths['width'], padding_lengths['channels'])
        else:
            pad_img = torch.zeros(padding_lengths['channels'],
                              padding_lengths['height'], padding_lengths['width'])
        pad_img[:img.shape[0], :img.shape[1], :img.shape[2]].copy_(img)
        return pad_img

    @overrides
    def empty_field(self):  # pylint: disable=no-self-use
        return ImageField(np.empty(self.image.shape), padding_value=self.padding_value)

    def __str__(self) -> str:
        return f"ImageField with shape: {self.image.shape}."


class MaskField(ImageField):
    """
    You may want to use a ``MaskField`` to store an image that should be treated as a label
    instead of an input.
    """

    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        np_img = self.image
        return torch.from_numpy(np_img).squeeze().float()

    def __str__(self) -> str:
        return f"MaskField with shape: {self.image.shape}."


class BoundingBoxField(Field[np.ndarray]):

    def __init__(self, coords: Sequence[float], padding_value: int = 0) -> None:
        self.coords = coords
        self.padding_value = padding_value

    @overrides
    def get_padding_lengths(self) -> Dict[str, int]:
        return {
            'coords': 4
        }

    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        return torch.tensor(self.coords, dtype=torch.float32)

    @overrides
    def empty_field(self):  # pylint: disable=no-self-use
        return BoundingBoxField([0, 0, 0, 0], padding_value=self.padding_value)

    def __str__(self) -> str:
        return f"BoundingBoxField with coords: {self.coords}"



import torch
import torch.nn as nn

from allencv.common.testing import ModelTestCase
from allencv.models.object_detection import RPN, FasterRCNN, PretrainedDetectronFasterRCNN
from allencv.modules.image_encoders import ResnetEncoder, FPN
from allencv.modules.im2vec_encoders import FlattenEncoder

from allencv.common.testing import AllenCvTestCase, ModelTestCase
from allencv.data.dataset_readers import ImageAnnotationReader
from allencv.models.object_detection import RPN
from allencv.modules.image_encoders import ResnetEncoder, FPN

from allennlp.modules import FeedForward


class TestFasterRCNN(ModelTestCase):

    def test_basic_experiment(self):
        data_directory = AllenCvTestCase.FIXTURES_ROOT / "data" / "image_annotation"
        self.set_up_model(AllenCvTestCase.FIXTURES_ROOT / 'object_detection' / 'faster_rcnn' / 'experiment.jsonnet',
                          data_directory)
        self.ensure_model_can_train_save_and_load(self.param_file, tolerance=1e-4)

    def test_pretrained_experiment(self):
        data_directory = AllenCvTestCase.FIXTURES_ROOT / "data" / "image_annotation"
        self.set_up_model(AllenCvTestCase.FIXTURES_ROOT / 'object_detection' / 'faster_rcnn' / 'pretrained_experiment.jsonnet',
                          data_directory)
        self.ensure_model_can_train_save_and_load(self.param_file, tolerance=1e-4)

    # def test_forward(self):
    #     batch_size = 3
    #     im = torch.randn(batch_size, 3, 224, 224)
    #     im_sizes = torch.tensor([im.shape[-2:] for _ in range(im.shape[0])]).view(3, -1)
    #     backbone = ResnetEncoder('resnet50')
    #     fpn_out_channels = 256
    #     fpn_backbone = FPN(backbone, fpn_out_channels)
    #     anchor_sizes = [32, 64, 128, 256, 512]
    #     anchor_strides = [4, 8, 16, 32, 64]
    #     fpn_pos_nms_top_n = 400
    #     rpn = RPN(fpn_backbone, anchor_sizes=anchor_sizes, anchor_strides=anchor_strides,
    #               fpn_post_nms_top_n=fpn_pos_nms_top_n)
    #     pooler_resolution = 7
    #     roi_num_features = pooler_resolution * pooler_resolution * fpn_out_channels
    #     feedforward = FeedForward(roi_num_features, num_layers=2, hidden_dims=[1024, 1024],
    #                               activations=nn.ReLU())
    #     encoder = FlattenEncoder(fpn_out_channels, pooler_resolution, pooler_resolution,
    #                              feedforward)
    #     num_labels = 4
    #     decoder_detections_per_image = 73
    #     frcnn = FasterRCNN(None, rpn, encoder, num_labels=num_labels,
    #                        decoder_detections_per_image=decoder_detections_per_image)
    #     out = frcnn.forward(im, im_sizes)
    #     assert list(out['class_logits'].shape) == [fpn_pos_nms_top_n, num_labels]
    #
    #     decoded = frcnn.decode(out)
    #     assert list(decoded['decoded'].shape) == [batch_size, decoder_detections_per_image, 4]

    # def test_pretrained(self):
    #     batch_size = 3
    #     im = torch.randn(batch_size, 3, 224, 224)
    #     im_sizes = torch.tensor([im.shape[-2:] for _ in range(im.shape[0])]).view(3, -1)
    #     anchor_sizes = [32, 64, 128, 256, 512]
    #     anchor_strides = [4, 8, 16, 32, 64]
    #     decoder_detections_per_image = 73
    #     rpn = PretrainedDetectronRPN(anchor_sizes=anchor_sizes, anchor_strides=anchor_strides)
    #     frcnn = PretrainedDetectronFasterRCNN(rpn,
    #                                           decoder_thresh=0.0,
    #                                           decoder_detections_per_image=decoder_detections_per_image)
    #     out = frcnn.forward(im, im_sizes)
    #     assert out['class_logits'].shape[1] == 81
    #     decoded = frcnn.decode(out)
    #     assert list(decoded['decoded'].shape) == [batch_size, decoder_detections_per_image, 4]





from allencv.common.testing import AllenCvTestCase, ModelTestCase
from allencv.data.dataset_readers import ImageAnnotationReader
from allencv.models.object_detection import RPN
from allencv.modules.image_encoders import ResnetEncoder, FPN


class TestRegionProposalNetwork(ModelTestCase):

    def test_basic_experiment(self):
        data_directory = AllenCvTestCase.FIXTURES_ROOT / "data" / "image_annotation"
        self.set_up_model(AllenCvTestCase.FIXTURES_ROOT / 'object_detection' / 'region_proposal_network' / 'experiment.jsonnet',
                          data_directory)
        self.ensure_model_can_train_save_and_load(self.param_file, tolerance=1e-4)

    def test_pretrained_experiment(self):
        data_directory = AllenCvTestCase.FIXTURES_ROOT / "data" / "image_annotation"
        self.set_up_model(AllenCvTestCase.FIXTURES_ROOT / 'object_detection' / 'region_proposal_network' / 'pretrained_experiment.jsonnet',
                          data_directory)
        self.ensure_model_can_train_save_and_load(self.param_file, tolerance=1e-4)




from typing import Tuple

from allencv.modules.encoder_base import _EncoderBase
from allennlp.common import Registrable


class Im2ImEncoder(_EncoderBase, Registrable):
    """
    A ``Im2ImEncoder`` is a ``Module`` that takes as input an image and returns a
    modified image. The height and width of the output image are variable depending on the
    dimensions of the input image.

    Input shape: ``(batch_size, input_channels, input_height, input_width)``;
    output shape: ``(batch_size, output_channels, output_height, output_width)``.
    """
    def get_input_channels(self) -> int:
        """
        Returns the number of input channels to this encoder.
        """
        raise NotImplementedError

    def get_output_channels(self) -> int:
        """
        Returns the number of output channels to this encoder.
        """
        raise NotImplementedError



from allencv.modules.image_encoders.image_encoder import ImageEncoder
from allencv.modules.image_encoders.resnet_encoder import ResnetEncoder
from allencv.modules.image_encoders.feature_pyramid_encoder import FPN



import numpy as np
from typing import List, Sequence

import torch
import torch.nn as nn

from allencv.nn import Upsample
from allencv.modules.image_decoders import ImageDecoder


@ImageDecoder.register("basic")
class BasicDecoder(ImageDecoder):
    """
    Simple decoder that takes several encoded images, each with the same number of channels,
    upscales them to a particular size, and sums them. Each image is upscaled by a
    factor of 2 at a time, with convolution stages in between, until it reaches the desired size.

    Parameters
    ----------
    input_scales: ``List[int]``
        What factor to scale each image by in the forward pass. Must be powers of 2.
    input_channels: ``int``, the number of channels in each of the encoded images.
    output_channels: ``int``, output channels in the decoded image.
    """

    def __init__(self,
                 input_scales: List[int],
                 input_channels: int,
                 output_channels: int) -> None:
        super(BasicDecoder, self).__init__()
        up_scales = [int(np.log2(x)) for x in input_scales]
        self._output_channels = output_channels
        self.upsampling_stages = nn.ModuleList()
        for n in up_scales:
            stage = Upsample(input_channels, self._output_channels, num_stages=n, scale_factor=2)
            self.upsampling_stages.append(stage)

    def forward(self,  # type: ignore
                images: Sequence[torch.Tensor]) -> torch.Tensor:
        # pylint: disable=arguments-differ
        upsampled_output = 0.
        for upstage, image in zip(self.upsampling_stages, images):
            upsampled_output += upstage.forward(image)
        return upsampled_output

    def get_output_channels(self):
        return self._output_channels




from overrides import overrides
from typing import Sequence, Union

import torch
import torch.nn as nn

from allennlp.nn import Activation

from allencv.nn import StdConv
from allencv.modules.im2im_encoders.im2im_encoder import Im2ImEncoder


@Im2ImEncoder.register("feedforward")
class FeedforwardEncoder(Im2ImEncoder):
    """
    A simple image encoder that passes an input through a series of convolutional layers,
    progressively halving the input size.

    Parameters
    ----------
    input_channels: ``int``
    num_layers: ``int``
        Number of convolutional transforms
    hidden_channels: ``Union[int, Sequence[int]]``
        The number of output channels for each of the convolutional stages.
    activations: ``str``
        The activation function to use after each convolutional stage.
    kernel_sizes: ``Union[int, Sequence[int]]``
        The kernel size for each of the convolutional stages.
    dropout: ``float``
        The amount of dropout for each convolutional stage.
    """

    def __init__(self,
                 input_channels: int,
                 num_layers: int,
                 hidden_channels: Union[int, Sequence[int]],
                 activations: str,
                 kernel_sizes: Union[int, Sequence[int]] = 3,
                 dropout: Union[float, Sequence[float]] = 0.0) -> None:
        super(FeedforwardEncoder, self).__init__()
        if not isinstance(hidden_channels, list):
            hidden_channels = [hidden_channels] * num_layers
        if not isinstance(activations, list):
            activations = [activations] * num_layers
        activations = [Activation.by_name(a)() for a in activations]
        if not isinstance(dropout, list):
            dropout = [dropout] * num_layers
        if not isinstance(kernel_sizes, list):
            kernel_sizes = [kernel_sizes] * num_layers
        self._activations = activations
        self._kernel_sizes = kernel_sizes
        input_channels = [input_channels] + hidden_channels[:-1]
        conv_layers = []
        for i, (layer_input_channel, layer_output_channel) in enumerate(zip(input_channels, hidden_channels)):
            conv_layers.append(StdConv(layer_input_channel, layer_output_channel,
                                       kernel_size=kernel_sizes[i],
                                       activation=activations[i],
                                       dropout=dropout[i]))
            conv_layers.append(StdConv(layer_output_channel, layer_output_channel, stride=2,
                                       kernel_size=kernel_sizes[i],
                                       activation=activations[i]))
        self._conv_layers = nn.ModuleList(conv_layers)
        self._output_channels = hidden_channels[-1]
        self.input_channels = input_channels[0]

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        # pylint: disable=arguments-differ
        output = inputs
        for layer in self._conv_layers:
            output = layer.forward(output)
        return output

    @overrides
    def get_input_channels(self) -> int:
        return self.input_channels

    @overrides
    def get_output_channels(self) -> int:
        return self._output_channels




from overrides import overrides
from typing import Sequence

import torch

from allennlp.common import Registrable

from allencv.modules.encoder_base import _EncoderBase


class ImageEncoder(_EncoderBase, Registrable):

    @overrides
    def forward(self,  # pylint: disable=arguments-differ
                image: torch.Tensor) -> Sequence[torch.Tensor]:
        raise NotImplementedError

    def get_input_channels(self) -> int:
        raise NotImplementedError

    def get_output_channels(self) -> Sequence[int]:
        raise NotImplementedError

    def get_output_scales(self) -> Sequence[int]:
        raise NotImplementedError


from typing import Dict, List, Tuple

from collections import OrderedDict
import logging
from overrides import overrides

import torch
import torch.nn as nn
import torch.nn.functional as F

from allennlp.data import Vocabulary
from allennlp.modules import FeedForward
from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator
from allennlp.training.metrics import Average

from torchvision.models.detection import _utils as det_utils
from torchvision.ops import boxes as box_ops
from torchvision.models.detection import fasterrcnn_resnet50_fpn
from torchvision.ops.poolers import MultiScaleRoIAlign
from torchvision.models.detection.roi_heads import fastrcnn_loss

from allencv.models.object_detection import utils as object_utils
from allencv.modules.im2vec_encoders import Im2VecEncoder, FlattenEncoder

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@Model.register("faster_rcnn")
class FasterRCNN(Model):
    """
    A Faster-RCNN model first detects objects in an image using a ``RegionProposalNetwork``.
    Faster-RCNN further separates those objects into classes and refines their bounding boxes.

    Parameters
    ----------
    vocab : ``Vocabulary``
    rpn: ``RPN``
        The region proposal network that detects initial objects.
    roi_feature_extractor: ``Im2VecEncoder``
        Maps each region of interest into a vector of features.
    num_labels: ``int``
        Number of object classe.
    label_namespace: ``str``
        Vocabulary namespace corresponding to labels. By default, we use the "labels" namespace.
    train_rpn: ``bool``
        Whether to include the RPN's loss in the overall loss.
    pooler_resolution: ``int``
        The ROI pooler will output images of this size.
    pooler_sampling_ratio: ``int``
        The sampling ratio for ROIAlign.
    decoder_thresh: ``float``
        The minimum score for an object proposal to make it in the final output.
    decoder_nms_thresh: ``float``
        Proposals that overlap by more than this number will be suppressed into one.
    decoder_detections_per_image: ``int``
        Number of object detections to be proposed per image.
    matcher_high_thresh: ``float``
        Any anchor boxes that overlap with target boxes with an IOU threshold less than
        this number will be considered background.
    matcher_low_thresh: ``float``
        Any anchor boxes that overlap with target boxes with an IOU threshold greater than
        this number will be considered foreground.
    allow_low_quality_matches: ``bool``
        Produce additional matches for predictions that have only low-quality matches.
    batch_size_per_image: ``int``
        The number of examples to include in the loss function for each image in each batch.
    balance_sampling_fraction: ``float``
        What fraction of each batch in the loss computation should be positive
        examples (foreground).
    class_agnostic_bbox_reg: ``bool``
        Whether to use a separate network for each class's bounding box refinement or a single
        network for all classes.
    """

    def __init__(self,
                 vocab: Vocabulary,
                 rpn: Model,
                 roi_feature_extractor: Im2VecEncoder,
                 num_labels: int = None,
                 label_namespace: str = "labels",
                 train_rpn: bool = False,
                 pooler_resolution: int = 7,
                 pooler_sampling_ratio: int = 2,
                 decoder_thresh: float = 0.1,
                 decoder_nms_thresh: float = 0.5,
                 decoder_detections_per_image: int = 100,
                 matcher_high_thresh: float = 0.5,
                 matcher_low_thresh: float = 0.5,
                 allow_low_quality_matches: bool = True,
                 batch_size_per_image: int = 64,
                 balance_sampling_fraction: float = 0.25,
                 class_agnostic_bbox_reg: bool = False,
                 initializer: InitializerApplicator = InitializerApplicator()) -> None:
        super(FasterRCNN, self).__init__(vocab)
        self.decoder_thresh = decoder_thresh
        self.decoder_nms_thresh = decoder_nms_thresh
        self.decoder_detections_per_image = decoder_detections_per_image
        self._train_rpn = train_rpn
        self.rpn = rpn
        if num_labels:
            self._num_labels = num_labels
        else:
            self._num_labels = vocab.get_vocab_size(namespace=label_namespace)
        self.feature_extractor = roi_feature_extractor
        num_bbox_reg_classes = 2 if class_agnostic_bbox_reg else self._num_labels
        representation_size = roi_feature_extractor.get_output_dim()
        self.cls_score = nn.Linear(representation_size, self._num_labels)
        self.bbox_pred = nn.Linear(representation_size, num_bbox_reg_classes * 4)

        self.proposal_matcher = det_utils.Matcher(
                high_threshold=matcher_high_thresh,
                low_threshold=matcher_low_thresh,
                allow_low_quality_matches=allow_low_quality_matches)
        self.sampler = det_utils.BalancedPositiveNegativeSampler(
                batch_size_per_image,
                positive_fraction=balance_sampling_fraction)
        self.box_coder = det_utils.BoxCoder(weights=(10., 10., 5., 5.))
        self.box_roi_pool = MultiScaleRoIAlign(
                featmap_names=[0, 1, 2, 3],
                output_size=pooler_resolution,
                sampling_ratio=pooler_sampling_ratio)
        self._loss_meters = {'roi_cls_loss': Average(), 'roi_reg_loss': Average(),
                             'rpn_cls_loss': Average(), 'rpn_reg_loss': Average()}
        initializer(self)

    def forward(self,
                image: torch.Tensor,  # (batch_size, c, h, w)
                image_sizes: torch.Tensor,  # (batch_size, 2)
                boxes: torch.Tensor = None,  # (batch_size, s, 4)
                box_classes: torch.Tensor = None):
        # pylint: disable=arguments-differ
        rpn_out = self.rpn.forward(image, image_sizes, boxes)
        rpn_out = self.rpn.decode(rpn_out)
        features: List[torch.Tensor] = rpn_out['features']

        proposals: List[torch.Tensor] = \
            object_utils.padded_tensor_to_tensor_list(rpn_out['proposals'])
        if boxes is not None:
            # subsample each set of proposals to a pre-specified number, e.g. 512
            with torch.no_grad():
                # TODO: fix single dimension case?
                gt_labels = [x.squeeze() for x in object_utils.padded_tensor_to_tensor_list(box_classes)]
                gt_boxes = object_utils.padded_tensor_to_tensor_list(boxes)
                matched_idxs, labels = self._assign_targets_to_proposals(
                        proposals, gt_boxes, gt_labels)
                # subsample
                sampled_pos_inds, sampled_neg_inds = self.sampler(labels)
                sampled_inds = []
                for pos_inds_img, neg_inds_img in zip(sampled_pos_inds, sampled_neg_inds):
                    img_sampled_inds = torch.nonzero(pos_inds_img | neg_inds_img).squeeze(1)
                    sampled_inds.append(img_sampled_inds)

                matched_gt_boxes = []
                num_images = len(proposals)
                for img_id in range(num_images):
                    img_sampled_inds = sampled_inds[img_id]
                    proposals[img_id] = proposals[img_id][img_sampled_inds]
                    labels[img_id] = labels[img_id][img_sampled_inds]
                    matched_idxs[img_id] = matched_idxs[img_id][img_sampled_inds]
                    matched_gt_boxes.append(gt_boxes[img_id][matched_idxs[img_id]])

                regression_targets = self.box_coder.encode(matched_gt_boxes, proposals)

        # roi pool and pass through feature extractor
        # (b, num_proposal_regions_in_batch, 14, 14)
        im_sizes = [(x[1].item(), x[0].item()) for x in image_sizes]
        _features = OrderedDict([(i, f) for i, f in enumerate(features)])
        pooled = self.box_roi_pool(_features, proposals, im_sizes)
        box_features = self.feature_extractor.forward(pooled)
        class_logits = self.cls_score(box_features)
        box_regression = self.bbox_pred(box_features)

        # pooler, combines logits from all images in the batch to one large batch
        # you would need to split the tensor by the number of proposals in each image to
        # recover the per-image logits
        out = {'class_logits': class_logits,
               'box_regression': box_regression,
               'proposals': proposals,
               'image_sizes': im_sizes}

        if boxes is not None:
            rpn_classifier_loss = rpn_out['loss_objectness']
            rpn_regression_loss = rpn_out['loss_rpn_box_reg']
            classifier_loss, regression_loss = fastrcnn_loss(
                    class_logits, box_regression, labels, regression_targets)
            loss = 0.003 * classifier_loss + regression_loss
            if self._train_rpn:
                loss += 0.1 * rpn_classifier_loss + rpn_regression_loss
            out['loss'] = loss
            self._loss_meters['rpn_cls_loss'](rpn_out['loss_objectness'].item())
            self._loss_meters['rpn_reg_loss'](rpn_out['loss_rpn_box_reg'].item())
            self._loss_meters['roi_cls_loss'](classifier_loss.item())
            self._loss_meters['roi_reg_loss'](regression_loss.item())
        return out

    @overrides
    def decode(self, output_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        # TODO: provide split class and regression tensors?
        boxes, scores, labels = self._postprocess_detections(
                output_dict['class_logits'],
                output_dict['box_regression'],
                output_dict['proposals'],
                output_dict['image_sizes'])
        output_dict['scores'] = object_utils.pad_tensors(scores)
        output_dict['labels'] = object_utils.pad_tensors(labels)
        output_dict['decoded'] = object_utils.pad_tensors(boxes)
        output_dict.pop("proposals")

        # convert label indices to classes
        all_predictions = output_dict['labels']
        all_predictions = all_predictions.cpu().data.numpy()
        if all_predictions.ndim == 2:
            predictions_list = [all_predictions[i] for i in range(all_predictions.shape[0])]
        else:
            predictions_list = [all_predictions]
        if self.vocab is not None:
            idx2token = self.vocab.get_index_to_token_vocabulary(namespace="labels")
            all_classes = []
            for predictions in predictions_list:
                all_classes.append([idx2token[x] for x in predictions])
            output_dict['class'] = all_classes
        # split_logits = class_logits.split([len(p) for p in sampled_proposals])
        # # [(13, 4), (17, 4)] -> (2, 17, 4)
        # split_logits = self.rpn._pad_tensors(split_logits)
        # split_box_regression = box_regression.split([len(p) for p in sampled_proposals])
        # split_box_regression = self.rpn._pad_tensors(split_box_regression)
        return output_dict

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = {k: v.get_metric(reset) for k, v in self._loss_meters.items()}
        return metrics

    def _postprocess_detections(self,
                                class_logits: torch.Tensor,
                                box_regression: torch.Tensor,
                                proposals: List[torch.Tensor],
                                image_shapes: List[Tuple[int, int]]):
        device = class_logits.device
        num_classes = class_logits.shape[-1]

        boxes_per_image = [len(boxes_in_image) for boxes_in_image in proposals]
        pred_boxes = self.box_coder.decode(box_regression, proposals)

        pred_scores = F.softmax(class_logits, -1)

        # split boxes and scores per image
        pred_boxes = pred_boxes.split(boxes_per_image, 0)
        pred_scores = pred_scores.split(boxes_per_image, 0)

        all_boxes = []
        all_scores = []
        all_labels = []
        for boxes, scores, image_shape in zip(pred_boxes, pred_scores, image_shapes):
            boxes = box_ops.clip_boxes_to_image(boxes, image_shape)

            # create labels for each prediction
            labels = torch.arange(num_classes, device=device)
            labels = labels.view(1, -1).expand_as(scores)

            # remove predictions with the background label
            boxes = boxes[:, 1:]
            scores = scores[:, 1:]
            labels = labels[:, 1:]

            # batch everything, by making every class prediction be a separate instance
            boxes = boxes.reshape(-1, 4)
            scores = scores.flatten()
            labels = labels.flatten()

            # remove low scoring boxes
            inds = torch.nonzero(scores > self.decoder_thresh).squeeze(1)
            boxes, scores, labels = boxes[inds], scores[inds], labels[inds]

            # non-maximum suppression, independently done per class
            keep = box_ops.batched_nms(boxes, scores, labels, self.decoder_nms_thresh)
            # keep only topk scoring predictions
            keep = keep[:self.decoder_detections_per_image]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            all_boxes.append(boxes)
            all_scores.append(scores)
            all_labels.append(labels)

        return all_boxes, all_scores, all_labels

    def _assign_targets_to_proposals(self,
                                     proposals: List[torch.Tensor],
                                     gt_boxes: List[torch.Tensor],
                                     gt_labels: List[torch.Tensor]):
        matched_idxs = []
        labels = []
        for proposals_in_image, gt_boxes_in_image, gt_labels_in_image in zip(proposals, gt_boxes, gt_labels):
            match_quality_matrix = box_ops.box_iou(gt_boxes_in_image, proposals_in_image)
            matched_idxs_in_image = self.proposal_matcher(match_quality_matrix)

            clamped_matched_idxs_in_image = matched_idxs_in_image.clamp(min=0)

            labels_in_image = gt_labels_in_image[clamped_matched_idxs_in_image]
            labels_in_image = labels_in_image.to(dtype=torch.int64)

            # Label background (below the low threshold)
            bg_inds = matched_idxs_in_image == self.proposal_matcher.BELOW_LOW_THRESHOLD
            labels_in_image[bg_inds] = 0

            # Label ignore proposals (between low and high thresholds)
            ignore_inds = matched_idxs_in_image == self.proposal_matcher.BETWEEN_THRESHOLDS
            labels_in_image[ignore_inds] = -1  # -1 is ignored by sampler

            matched_idxs.append(clamped_matched_idxs_in_image)
            labels.append(labels_in_image)
        return matched_idxs, labels


@Model.register("pretrained_detectron_faster_rcnn")
class PretrainedDetectronFasterRCNN(FasterRCNN):

    CATEGORIES = [
            'unlabeled',
            'person',
            'bicycle',
            'car',
            'motorcycle',
            'airplane',
            'bus',
            'train',
            'truck',
            'boat',
            'traffic light',
            'fire hydrant',
            'street sign',
            'stop sign',
            'parking meter',
            'bench',
            'bird',
            'cat',
            'dog',
            'horse',
            'sheep',
            'cow',
            'elephant',
            'bear',
            'zebra',
            'giraffe',
            'hat',
            'backpack',
            'umbrella',
            'shoe',
            'eye glasses',
            'handbag',
            'tie',
            'suitcase',
            'frisbee',
            'skis',
            'snowboard',
            'sports ball',
            'kite',
            'baseball bat',
            'baseball glove',
            'skateboard',
            'surfboard',
            'tennis racket',
            'bottle',
            'plate',
            'wine glass',
            'cup',
            'fork',
            'knife',
            'spoon',
            'bowl',
            'banana',
            'apple',
            'sandwich',
            'orange',
            'broccoli',
            'carrot',
            'hot dog',
            'pizza',
            'donut',
            'cake',
            'chair',
            'couch',
            'potted plant',
            'bed',
            'mirror',
            'dining table',
            'window',
            'desk',
            'toilet',
            'door',
            'tv',
            'laptop',
            'mouse',
            'remote',
            'keyboard',
            'cell phone',
            'microwave',
            'oven',
            'toaster',
            'sink',
            'refrigerator',
            'blender',
            'book',
            'clock',
            'vase',
            'scissors',
            'teddy bear',
            'hair drier',
            'toothbrush']

    def __init__(self,
                 rpn: Model,
                 train_rpn: bool = False,
                 pooler_sampling_ratio: int = 2,
                 decoder_thresh: float = 0.1,
                 decoder_nms_thresh: float = 0.5,
                 decoder_detections_per_image: int = 100,
                 matcher_high_thresh: float = 0.5,
                 matcher_low_thresh: float = 0.5,
                 allow_low_quality_matches: bool = True,
                 batch_size_per_image: int = 64,
                 balance_sampling_fraction: float = 0.25):
        feedforward = FeedForward(7 * 7 * 256, 2, [1024, 1024], nn.ReLU())
        encoder = FlattenEncoder(256, 7, 7, feedforward)
        vocab = Vocabulary({'labels': {k: 1 for k in PretrainedDetectronFasterRCNN.CATEGORIES}})
        super(PretrainedDetectronFasterRCNN, self)\
            .__init__(vocab, rpn, encoder,
                      pooler_resolution=7,
                      train_rpn=train_rpn,
                      pooler_sampling_ratio=pooler_sampling_ratio,
                      matcher_low_thresh=matcher_low_thresh,
                      matcher_high_thresh=matcher_high_thresh,
                      decoder_thresh=decoder_thresh,
                      decoder_nms_thresh=decoder_nms_thresh,
                      decoder_detections_per_image=decoder_detections_per_image,
                      allow_low_quality_matches=allow_low_quality_matches,
                      batch_size_per_image=batch_size_per_image,
                      balance_sampling_fraction=balance_sampling_fraction,
                      label_namespace='labels',
                      class_agnostic_bbox_reg=False)
        frcnn = fasterrcnn_resnet50_fpn(pretrained=True, pretrained_backbone=True)
        self.cls_score.load_state_dict(frcnn.roi_heads.box_predictor.cls_score.state_dict())
        self.bbox_pred.load_state_dict(frcnn.roi_heads.box_predictor.bbox_pred.state_dict())

        # pylint: disable = protected-access
        feedforward._linear_layers[0].load_state_dict(frcnn.roi_heads.box_head.fc6.state_dict())
        feedforward._linear_layers[1].load_state_dict(frcnn.roi_heads.box_head.fc7.state_dict())



# pylint: disable=no-self-use,invalid-name
import numpy as np
import torch

from allencv.common.testing.test_case import AllenCvTestCase
from allencv.data.fields import BoundingBoxField, ImageField, MaskField


class TestImageField(AllenCvTestCase):

    def test_image_field(self):
        im = np.random.randn(3, 8, 10)
        im_field = ImageField(im)
        assert list(im_field.as_tensor(im_field.get_padding_lengths()).shape) == [3, 8, 10]
        padding = im_field.get_padding_lengths()
        assert padding['channels'] == 3
        assert padding['height'] == 8
        assert padding['width'] == 10

        im = np.random.randn(8, 10, 3)
        im_field = ImageField(im)
        assert list(im_field.as_tensor(im_field.get_padding_lengths()).shape) == [8, 10, 3]

    def test_bounding_box_field(self):
        bbox_field = BoundingBoxField([0, 0, 10, 20])
        padding = bbox_field.get_padding_lengths()
        assert list(bbox_field.as_tensor(padding).shape) == [4]
        self.assertTrue(torch.all(bbox_field.empty_field().as_tensor(padding) == 0.).item())

    def test_mask_field(self):
        im = np.random.randn(3, 8, 10)
        mask_field = MaskField(im)
        assert list(mask_field.as_tensor(mask_field.get_padding_lengths()).shape) == [3, 8, 10]

        im = np.random.randn(1, 8, 10)
        mask_field = MaskField(im)
        assert list(mask_field.as_tensor(mask_field.get_padding_lengths()).shape) == [8, 10]

        im = np.random.randn(8, 10, 1)
        mask_field = MaskField(im, channels_first=False)
        assert list(mask_field.as_tensor(mask_field.get_padding_lengths()).shape) == [8, 10]

        im = np.random.randn(8, 10)
        mask_field = MaskField(im)
        assert list(mask_field.as_tensor(mask_field.get_padding_lengths()).shape) == [8, 10]

    def test_padding(self):
        im = np.random.randn(3, 8, 10)
        im_field = ImageField(im)
        im_tensor = im_field.as_tensor({'channels': 4, 'width': 11, 'height': 37})
        assert tuple(im_tensor.shape) == (4, 37, 11)
        self.assertTrue(torch.all(im_tensor[-1] == 0.).item())
        self.assertTrue(torch.all(im_tensor[:, 8:, :] == 0.).item())
        self.assertTrue(torch.all(im_tensor[:, :, 10:] == 0.).item())



# pylint: disable=no-self-use,invalid-name
from allencv.common.testing.test_case import AllenCvTestCase
from allencv.data.dataset_readers import ImageClassificationDirectory
from allencv.data.fields import ImageField

from allennlp.data.fields import LabelField


class TestImageClassificationDirectoryReader(AllenCvTestCase):

    def test_reader_output(self):
        reader = ImageClassificationDirectory()
        instances = reader.read(self.FIXTURES_ROOT / "data" / "image_classification")
        assert(isinstance(instances, list))
        assert len(instances) == 2
        fields = instances[0].fields
        assert isinstance(fields['image'], ImageField)
        assert isinstance(fields['label'], LabelField)
        assert set([inst.fields['label'].label for inst in instances]) == {'Cat', 'Dog'}





import logging
import torch
from typing import List, Tuple

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def sample_balanced_classes(
        matched_idxs: List[torch.tensor],
        batch_size_per_image: int,
        positive_fraction: float = 0.5) -> Tuple[List[torch.tensor], List[torch.tensor]]:
    """
    Arguments:
        matched idxs: list of tensors containing -1, 0 or positive values.
            Each tensor corresponds to a specific image.
            -1 values are ignored, 0 are considered as negatives and > 0 as
            positives.

    Returns:
        pos_idx (list[tensor])
        neg_idx (list[tensor])

    Returns two lists of binary masks for each image.
    The first list contains the positive elements that were selected,
    and the second list the negative example.
    """

    pos_idx = []
    neg_idx = []
    for matched_idxs_per_image in matched_idxs:
        positive = torch.nonzero(matched_idxs_per_image >= 1).squeeze(1)
        negative = torch.nonzero(matched_idxs_per_image == 0).squeeze(1)

        num_pos = int(batch_size_per_image * positive_fraction)
        # protect against not enough positive examples
        num_pos = min(positive.numel(), num_pos)
        num_neg = batch_size_per_image - num_pos
        # protect against not enough negative examples
        num_neg = min(negative.numel(), num_neg)

        if num_pos < 0.5 * batch_size_per_image * positive_fraction:
            logger.warning(f"Only found {num_pos / (num_neg + num_pos)} positive fraction.")

        # randomly select positive and negative examples
        perm1 = torch.randperm(positive.numel(), device=positive.device)[:num_pos]
        perm2 = torch.randperm(negative.numel(), device=negative.device)[:num_neg]

        pos_idx_per_image = positive[perm1]
        neg_idx_per_image = negative[perm2]

        # create binary mask from indices
        pos_idx_per_image_mask = torch.zeros_like(
            matched_idxs_per_image, dtype=torch.uint8
        )
        neg_idx_per_image_mask = torch.zeros_like(
            matched_idxs_per_image, dtype=torch.uint8
        )
        pos_idx_per_image_mask[pos_idx_per_image] = 1
        neg_idx_per_image_mask[neg_idx_per_image] = 1

        pos_idx.append(pos_idx_per_image_mask)
        neg_idx.append(neg_idx_per_image_mask)

    return pos_idx, neg_idx



# pylint: disable=no-self-use,invalid-name
from allencv.common.testing.test_case import AllenCvTestCase
from allencv.data.dataset_readers import PairedImageReader
from allencv.data.fields import ImageField, MaskField


class TestImageMaskReader(AllenCvTestCase):

    def test_reader_output(self):
        reader = PairedImageReader(mask_ext=".png")
        instances = reader.read(TestImageMaskReader.FIXTURES_ROOT / "data" / "image_mask_reader")
        assert(isinstance(instances, list))
        assert len(instances) == 6
        fields = instances[0].fields
        assert isinstance(fields['image'], ImageField)
        assert isinstance(fields['mask'], MaskField)

    def test_lazy(self):
        reader = PairedImageReader(mask_ext=".png", lazy=True)
        instances = reader.read(TestImageMaskReader.FIXTURES_ROOT / "data" / "image_mask_reader")
        assert not isinstance(instances, list)
        assert len(list(instances)) == 6
        fields = list(instances)[0].fields
        assert isinstance(fields['image'], ImageField)
        assert isinstance(fields['mask'], MaskField)

    def test_mask_extension_default(self):
        reader = PairedImageReader(mask_dir="images")
        instances = reader.read(TestImageMaskReader.FIXTURES_ROOT / "data" / "image_mask_reader")
        assert len(instances) == 6





import torch

from allencv.common.testing import AllenCvTestCase

from allencv.modules.image_decoders import BasicDecoder


class TestBasicDecoder(AllenCvTestCase):

    def test_decoder_shapes(self):
        in_channels = 256
        out_channels = 128
        scales = [4, 2, 1]
        decoder = BasicDecoder(scales, in_channels, out_channels)
        input_sizes = [16, 32, 64]
        inputs = [torch.randn(1, in_channels, size, size) for size in input_sizes]
        out = decoder.forward(inputs)
        assert out.shape[-2:] == (64, 64)
        assert out.shape[1] == out_channels
        assert decoder.get_output_channels() == out_channels




# pylint: disable=no-self-use,invalid-name
import torch

from allennlp.common.testing import AllenNlpTestCase

from allencv.modules.image_encoders import ResnetEncoder, FPN


class TestFeaturePyramid(AllenNlpTestCase):

    def test_encoder_shapes(self):
        image = torch.randn(1, 3, 224, 224)
        backbone = ResnetEncoder(resnet_model='resnet101')
        output_channels = 128
        fpn = FPN(backbone, output_channels)
        assert fpn.get_output_scales() == [4, 8, 16, 32]
        assert fpn.get_input_channels() == backbone.get_input_channels()
        assert fpn.get_output_channels() == [output_channels] * 5
        encoded = fpn.forward(image)
        assert len(encoded) == 5
        assert [im.shape[-2] for im in encoded] == [56, 28, 14, 7, 4]
        assert [im.shape[-1] for im in encoded] == [56, 28, 14, 7, 4]
        assert all([im.shape[1] == output_channels for im in encoded])
        image = torch.randn(1, 3, 500, 500)
        encoded = fpn.forward(image)
        assert [im.shape[-2] for im in encoded] == [125, 63, 32, 16, 8]
        assert [im.shape[-1] for im in encoded] == [125, 63, 32, 16, 8]





from typing import List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from allennlp.nn import Activation


class StdConv(nn.Module):

    def __init__(self,
                 nin: int,
                 nout: int,
                 kernel_size: int = 3,
                 activation: Activation = nn.ReLU(),
                 stride: int = 1,
                 padding: int = 1,
                 dilation: int = 1,
                 dropout: float = 0.1):
        super(StdConv, self).__init__()
        self.conv = nn.Conv2d(nin, nout, kernel_size, stride=stride, padding=padding, dilation=dilation)
        self.bn = nn.BatchNorm2d(nout)
        self.drop = nn.Dropout(dropout)
        self.activation = activation

    def forward(self, x):
        return self.drop(self.bn(self.activation(self.conv(x))))


class Upsample(nn.Module):

    def __init__(self, in_channels: int, out_channels: int, num_stages: int, scale_factor: int = 2):
        super(Upsample, self).__init__()
        self.scale_factor = scale_factor
        self.convs = [StdConv(in_channels, out_channels, padding=1)] + \
                [StdConv(out_channels, out_channels, padding=1) for _ in range(num_stages - 1)]
        self.convs = nn.ModuleList(self.convs)
        self.num_stages = num_stages

    def forward(self, inputs: torch.Tensor):
        out = inputs
        for conv in self.convs:
            out = conv.forward(out)
            if self.num_stages != 0:
                out = F.interpolate(out, scale_factor=self.scale_factor)
        return out


class RPNHead(nn.Module):
    def __init__(self, in_channels, num_anchors):
        super(RPNHead, self).__init__()
        self.conv = nn.Conv2d(
            in_channels, in_channels, kernel_size=3, stride=1, padding=1
        )
        self.cls_logits = nn.Conv2d(in_channels, num_anchors, kernel_size=1, stride=1)
        self.bbox_pred = nn.Conv2d(
            in_channels, num_anchors * 4, kernel_size=1, stride=1
        )

    def forward(self,
                features: List[torch.Tensor]) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        objectness: List[torch.Tensor] = []
        rpn_box_regression: List[torch.Tensor] = []
        for feature in features:
            t = F.relu(self.conv(feature))
            objectness.append(self.cls_logits(t))
            rpn_box_regression.append(self.bbox_pred(t))
        return objectness, rpn_box_regression



from overrides import overrides
from typing import Sequence

import torch

from allennlp.common import Registrable

from allencv.modules.decoder_base import _DecoderBase


class ImageDecoder(_DecoderBase, Registrable):

    @overrides
    def forward(self,  # pylint: disable=arguments-differ
                images: Sequence[torch.Tensor]) -> torch.Tensor:
        raise NotImplementedError

    def get_output_channels(self) -> int:
        raise NotImplementedError


import numpy as np
from typing import List

from allencv.data.transforms.image_transform import ImageTransform
from allennlp.data.dataset_readers import DatasetReader

import albumentations as aug


class ImageDatasetReader(DatasetReader):
    """
    A dataset reader that includes an augmentation method. This method will automatically
    identify all images and bounding boxes which may need to be transformed.

    Parameters
    ----------
    augmentation: ``ImageTransform``s to be applied in order to the data.
    bbox_format: ``str``, optional, (default='pascal_voc')
        How to interpret bounding box coordinates: either xyxy ('pascal_voc') or xywh ('coco')
    """

    def __init__(self,
                 augmentation: List[ImageTransform] = None,
                 bbox_format: str = 'pascal_voc',
                 lazy: bool = False):
        super(ImageDatasetReader, self).__init__(lazy)
        if augmentation is None:
            _transforms = []
        else:
            _transforms = [a.transform for a in augmentation]
        self.augmentation = aug.Compose(_transforms,
                                        bbox_params={'format': bbox_format,
                                                     'label_fields': ['category_id']})

    def augment(self,
                 image: np.ndarray,
                 masks: List[np.ndarray] = list(),
                 boxes: List[np.ndarray] = list()):
        augmented = self.augmentation(image=image, masks=masks,
                                      bboxes=boxes, category_id=[0]*len(boxes))
        return augmented['image'], augmented['masks'], augmented['bboxes']



from allencv.data.dataset_readers.paired_image_reader import PairedImageReader
from allencv.data.dataset_readers.image_classification_directory import ImageClassificationDirectory
from allencv.data.dataset_readers.image_annotation_reader import ImageAnnotationReader



from typing import List

import torch


def pad_tensors(tensors: List[torch.Tensor]):
    max_proposals = max([x.shape[0] for x in tensors])
    pad_shape = (len(tensors), max_proposals) + (
    () if tensors[0].dim() <= 1 else tensors[0].shape[1:])
    padded = torch.zeros(pad_shape, device=tensors[0].device)
    for i, t in enumerate(tensors):
        padded[i, :t.shape[0], ...] = t
    return padded


def padded_tensor_to_tensor_list(padded_tensor: torch.Tensor) -> List[torch.Tensor]:
    box_list = []
    for image_boxes in padded_tensor:
        # remove boxes that are all zeros (padding)
        mask = torch.any(image_boxes != 0., dim=-1)
        box_list.append(image_boxes[mask])
    return box_list




import logging
from overrides import overrides
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator
from allennlp.training.metrics import CategoricalAccuracy

from allencv.common import util
from allencv.modules.image_encoders import ImageEncoder
from allencv.modules.image_decoders import ImageDecoder

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@Model.register("semantic_segmentation")
class SemanticSegmentationModel(Model):
    """
    This model uses an ``ImageEncoder`` to extract features from an image at one or more
    scales, and then decodes those images to a single feature map, which is then upscaled
    back to the original size. It provides pixel level classifications from a known subset
    possible class labels.
    """
    def __init__(self,
                 encoder: ImageEncoder,
                 decoder: ImageDecoder,
                 num_classes: int,
                 batch_size_per_image: int = None,
                 sample_positive_fraction: float = 0.5,
                 initializer: InitializerApplicator = InitializerApplicator()) -> None:
        super(SemanticSegmentationModel, self).__init__(None)
        self._batch_size_per_image = batch_size_per_image
        self._sample_positive_fraction = sample_positive_fraction
        # TODO: use vocab to get label namespace?
        self._encoder = encoder
        self.num_classes = num_classes
        self._decoder = decoder
        self.final_conv = nn.Conv2d(in_channels=decoder.get_output_channels(),
                                    out_channels=self.num_classes,
                                    kernel_size=1)
        self._accuracy = CategoricalAccuracy()
        self._loss = nn.CrossEntropyLoss()
        initializer(self)

    def forward(self,  # type: ignore
                image: torch.FloatTensor,
                mask: torch.LongTensor = None) -> Dict[str, torch.Tensor]:
        # pylint: disable=arguments-differ

        encoded_images = self._encoder.forward(image)
        decoded = self._decoder.forward(encoded_images)
        decoded = self.final_conv(decoded)
        # scale image back to label size if possible, otherwise back to original image size
        if mask is not None:
            scale_factor = mask.shape[-2] // decoded.shape[-2]
        else:
            scale_factor = image.shape[-2] // decoded.shape[-2]

        logits = F.interpolate(decoded, scale_factor=scale_factor, mode='bilinear')

        probs = F.softmax(logits, dim=1)

        output_dict = {"logits": logits, "probs": probs}
        if mask is not None:
            flattened_logits = logits.permute(0, 2, 3, 1).contiguous().view(-1, self.num_classes)
            flattened_mask = mask.view(-1).long()
            # TODO: is it ok to silently filter labels that are out of range?
            ignore = flattened_mask >= self.num_classes
            flattened_mask[ignore] = -1
            # balance the positive and negative samples in each batch
            if self._batch_size_per_image is None:
                batch_size = image.shape[-2] * image.shape[-1]
            else:
                batch_size = self._batch_size_per_image
            neg_idxs, pos_idxs = util.sample_balanced_classes([flattened_mask],
                                                              batch_size * logits.shape[0],
                                                              self._sample_positive_fraction)
            sampled_pos_inds = torch.nonzero(torch.cat(neg_idxs, dim=0)).squeeze(1)
            sampled_neg_inds = torch.nonzero(torch.cat(pos_idxs, dim=0)).squeeze(1)

            sampled_inds = torch.cat([sampled_pos_inds, sampled_neg_inds], dim=0)
            # TODO: don't sample when evaluating
            if self.training:
                loss = self._loss(flattened_logits[sampled_inds, :], flattened_mask[sampled_inds])
            else:
                loss = self._loss(flattened_logits[~ignore, :], flattened_mask[~ignore])
            output_dict["loss"] = loss
            self._accuracy(flattened_logits[~ignore, :], flattened_mask[~ignore])
        return output_dict

    @overrides
    def decode(self, output_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        output_dict['predicted_mask'] = output_dict['probs'].argmax(dim=1)
        return output_dict

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = {'accuracy': self._accuracy.get_metric(reset)}
        return metrics



import json
import numpy as np
from overrides import overrides
from pathlib import Path
from PIL import Image

from typing import Dict, Iterable, List
import logging

from allencv.data.dataset_readers.image_dataset_reader import ImageDatasetReader
from allencv.data.fields.image_field import ImageField, BoundingBoxField
from allencv.data.transforms.image_transform import ImageTransform

from allennlp.data.instance import Instance
from allennlp.data.fields import ArrayField, Field, LabelField, ListField
from allennlp.data.dataset_readers import DatasetReader

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@DatasetReader.register("image_annotation")
class ImageAnnotationReader(ImageDatasetReader):
    """
    Read pairs of images and annotations. Expects annotations in the Pascal VOC format.

    data_path
    |-- images
    |   |-- 00001.jpg
    |   |-- 00002.jpg
    |   |-- ...
    |-- annotations
    |   |-- 00001.json
    |   |-- 00002.json
    |   |-- ...

    Parameters
    ----------
    augmentation: ``list``, required
        List of image augmentations to apply to each pair.
    image_dir: ``str``, optional, (default='images')
        Name of the subdirectory which contains image inputs.
    annotation_dir: ``str``, optional, (default='images')
    annotation_ext: ``str``, optional
        File extension for the annotation images.
    """
    def __init__(self,
                 augmentation: List[ImageTransform] = None,
                 image_dir: str = "images",
                 annotation_dir: str = "annotations",
                 annotation_ext: str = ".json",
                 lazy: bool = False) -> None:
        super(ImageAnnotationReader, self).__init__(augmentation, lazy=lazy)
        self._image_dir = image_dir
        self._annotation_dir = annotation_dir
        self._annotation_ext = annotation_ext
        self.lazy = lazy

    def _read(self, file_path: str) -> Iterable[Instance]:
        file_path = Path(file_path)

        for image_file in (file_path / self._image_dir).iterdir():
            img_name = image_file.stem
            img = Image.open(image_file)
            annotation_ext = self._annotation_ext
            annotation_file = file_path / self._annotation_dir / (img_name + annotation_ext)
            if not annotation_file.exists():
                label_boxes = None
                label_classes = None
            else:
                with open(annotation_file, 'r') as f:
                    annotation = json.load(f)
                label_boxes: List[List[float]] = self._parse_annotation(annotation)
                label_classes = [1] * len(label_boxes)  # TODO: fix this to read actual classes
            sample = img.convert('RGB')
            yield self.text_to_instance(np.array(sample), label_boxes, label_classes)

    @staticmethod
    def _parse_annotation(annotation) -> List[List[float]]:
        boxes = []
        for att in annotation['regions']:
            box = att['shape_attributes']
            x1, y1 = box['x'], box['y']
            w, h = box['width'], box['height']
            x2, y2 = x1 + w, y1 + h
            boxes.append([x1, y1, x2, y2])
        return boxes

    @overrides
    def text_to_instance(self,
                         image: np.ndarray,
                         label_box: List[List[float]] = None,
                         label_class: List[str] = None) -> Instance:
        if label_box is not None:
            img, _, label_box = self.augment(image, boxes=[np.array(b) for b in label_box])
        else:
            img, _, _ = self.augment(image)
        h, w, c = img.shape
        fields: Dict[str, Field] = {}
        fields['image'] = ImageField(img.transpose(2, 0, 1), channels_first=False)
        fields['image_sizes'] = ArrayField(np.array([w, h]))
        if label_box is not None:
            box_fields = [BoundingBoxField(x) for x in label_box]
            fields['boxes'] = ListField(box_fields)
            fields['box_classes'] = ListField([LabelField(idx, skip_indexing=True)
                                               for idx in label_class])
        return Instance(fields)



import torch
import torch.nn as nn

from allencv.models.object_detection import FasterRCNN
from allencv.modules.im2vec_encoders import FlattenEncoder

from allencv.common.testing import AllenCvTestCase
from allencv.data.dataset_readers import ImageAnnotationReader
from allencv.predictors import ImagePredictor
from allencv.models.object_detection import RPN
from allencv.modules.image_encoders import ResnetEncoder, FPN

from allennlp.modules import FeedForward


class TestFasterRCNN(AllenCvTestCase):

    def test_predictor(self):
        batch_size = 3
        im = torch.randn(batch_size, 3, 224, 224)
        im_sizes = torch.tensor([im.shape[-2:] for _ in range(im.shape[0])]).view(3, -1)
        backbone = ResnetEncoder('resnet18')
        fpn_out_channels = 256
        fpn_backbone = FPN(backbone, fpn_out_channels)
        anchor_sizes = [[32], [64], [128], [256], [512]]
        anchor_strides = [[0.5, 1.0, 2.0], [0.5, 1.0, 2.0], [0.5, 1.0, 2.0],
                          [0.5, 1.0, 2.0], [0.5, 1.0, 2.0]]
        fpn_pos_nms_top_n = 400
        rpn = RPN(fpn_backbone,
                  anchor_sizes=anchor_sizes,
                  anchor_strides=anchor_strides,
                  fpn_post_nms_top_n=fpn_pos_nms_top_n,
                  match_thresh_high=0.001,
                  match_thresh_low=0.0,
                  batch_size_per_image=10000000)
        pooler_resolution = 7
        roi_num_features = pooler_resolution * pooler_resolution * fpn_out_channels
        feedforward = FeedForward(roi_num_features, num_layers=2, hidden_dims=[32, 32],
                                  activations=nn.ReLU())
        encoder = FlattenEncoder(fpn_out_channels, pooler_resolution, pooler_resolution,
                                 feedforward)
        frcnn = FasterRCNN(None, rpn, encoder,
                           num_labels=5,
                           decoder_thresh=0.0,
                           decoder_nms_thresh=0.01,
                           batch_size_per_image=1000000)
        reader = ImageAnnotationReader()
        predictor = ImagePredictor(frcnn, reader)
        predicted = predictor.predict(AllenCvTestCase.FIXTURES_ROOT / "data" / "image_annotation" / "images" / "00001.jpg")
        assert len(predicted['scores']) == len(predicted['labels'])



import numpy as np
from overrides import overrides
from PIL import Image

from allennlp.common.util import JsonDict
from allennlp.data import Instance
from allennlp.predictors import Predictor


@Predictor.register("default_image")
class ImagePredictor(Predictor):

    def predict(self, image_path: str) -> JsonDict:
        return self.predict_json({"image_path": image_path})

    @overrides
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        if "image_path" in json_dict:
            image_path = json_dict["image_path"]
            img = np.array(Image.open(image_path).convert('RGB'))
        elif "image" in json_dict:
            img = np.array(json_dict['image']).reshape(json_dict['image_shape']).astype(np.float64)
        return self._dataset_reader.text_to_instance(img)



import numpy as np
from overrides import overrides
from pathlib import Path
from PIL import Image

from typing import Dict, Iterable, List
import logging

from allencv.data.dataset_readers.image_dataset_reader import ImageDatasetReader
from allencv.data.fields.image_field import ImageField, MaskField
from allencv.data.transforms.image_transform import ImageTransform

from allennlp.data.instance import Instance
from allennlp.data.fields import Field
from allennlp.data.dataset_readers import DatasetReader

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@DatasetReader.register("paired_image")
class PairedImageReader(ImageDatasetReader):
    """
    Read pairs of images as input and label.

    data_path
    |-- images
    |   |-- 00001.jpg
    |   |-- 00002.jpg
    |   |-- ...
    |-- masks
    |   |-- 00001.png
    |   |-- 00002.png
    |   |-- ...

    Parameters
    ----------
    augmentation: ``list``, required
        List of image augmentations to apply to each pair.
    image_dir: ``str``, optional, (default='images')
        Name of the subdirectory which contains image inputs.
    mask_dir: ``str``, optional, (default='images')
    mask_ext: ``str``, optional
        File extension for the mask images.
    """
    def __init__(self,
                 augmentation: List[ImageTransform] = None,
                 image_dir: str = "images",
                 mask_dir: str = "masks",
                 mask_ext: str = None,
                 lazy: bool = False) -> None:
        super(PairedImageReader, self).__init__(augmentation, lazy)
        self._image_dir = image_dir
        self._mask_dir = mask_dir
        self._mask_ext = mask_ext
        self.lazy = lazy

    def _read(self, file_path: str) -> Iterable[Instance]:
        file_path = Path(file_path)
        for image_file in (file_path / self._image_dir).iterdir():
            img_name = image_file.stem
            img = Image.open(image_file)
            mask_ext = self._mask_ext if self._mask_ext is not None else image_file.suffix
            mask_file = file_path / self._mask_dir / (img_name + mask_ext)
            mask = Image.open(mask_file)
            sample = img.convert('RGB')
            yield self.text_to_instance(np.array(sample), mask)

    @overrides
    def text_to_instance(self, image: np.ndarray, mask: np.ndarray = None) -> Instance:
        if mask is not None:
            img, masks, _ = self.augment(image, masks=[np.array(mask)])
            mask = masks[0]
        else:
            img, _, _ = self.augment(image)
        fields: Dict[str, Field] = {}
        fields['image'] = ImageField(img.transpose(2, 0, 1), channels_first=False)
        if mask is not None:
            if len(mask.shape) == 2:
                mask = mask[:, :, np.newaxis]
            fields['mask'] = MaskField(mask, channels_first=False)
        return Instance(fields)



# pylint: disable=no-self-use,invalid-name
import torch

from allennlp.common.testing import AllenNlpTestCase

from allencv.modules.image_encoders import ResnetEncoder


class TestResnetEncoder(AllenNlpTestCase):

    def test_encoder_shapes(self):
        image = torch.randn(1, 3, 224, 224)
        encoder = ResnetEncoder(resnet_model='resnet101')
        assert encoder.get_output_scales() == [4, 8, 16, 32]
        assert encoder.get_input_channels() == 3
        assert encoder.get_output_channels() == [256, 512, 1024, 2048]
        encoded = encoder.forward(image)
        assert len(encoded) == 4
        assert [im.shape[-2] for im in encoded] == [56, 28, 14, 7]
        assert [im.shape[-1] for im in encoded] == [56, 28, 14, 7]
        assert [im.shape[1] for im in encoded] == encoder.get_output_channels()
        image = torch.randn(1, 3, 500, 500)
        encoded = encoder.forward(image)
        assert [im.shape[-2] for im in encoded] == [125, 63, 32, 16]
        assert [im.shape[-1] for im in encoded] == [125, 63, 32, 16]





from typing import Dict

from overrides import overrides
import torch

from allennlp.data import Vocabulary
from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator
from allennlp.training.metrics import CategoricalAccuracy

from allencv.modules.im2im_encoders.im2im_encoder import Im2ImEncoder
from allencv.modules.im2vec_encoders.im2vec_encoder import Im2VecEncoder


@Model.register("basic_image_classifier")
class BasicImageClassifier(Model):
    """
    This ``Model`` implements a basic image classifier. Optionally encode the input image
    with a ``Im2ImEncoder`` and then encode the image with a ``Im2VecEncoder``. That output
    is passed to a linear classification layer, which projects into the label space.

    Parameters
    ----------
    vocab : ``Vocabulary``
    im2im_encoder : ``Im2ImEncoder``, optional (default=``None``)
        Optional image encoder layer for the input image.
    im2vec_encoder : ``Im2VecEncoder``
        Required Im2Vec encoder layer.
    dropout : ``float``, optional (default = ``None``)
        Dropout percentage to use.
    num_labels: ``int``, optional (default = ``None``)
        Number of labels to project to in classification layer. By default, the classification
        layer will project to the size of the vocabulary namespace corresponding to labels.
    label_namespace: ``str``, optional (default = "labels")
        Vocabulary namespace corresponding to labels. By default, we use the "labels" namespace.
    initializer : ``InitializerApplicator``, optional (default=``InitializerApplicator()``)
        If provided, will be used to initialize the model parameters.
    """
    def __init__(self,
                 vocab: Vocabulary,
                 im2vec_encoder: Im2VecEncoder,
                 im2im_encoder: Im2ImEncoder = None,
                 dropout: float = None,
                 num_labels: int = None,
                 label_namespace: str = "labels",
                 initializer: InitializerApplicator = InitializerApplicator()) -> None:

        super(BasicImageClassifier, self).__init__(vocab)
        self._im2vec_encoder = im2vec_encoder
        if im2im_encoder:
            self._im2im_encoder = im2im_encoder
        else:
            self._im2im_encoder = None
        self._classifier_input_dim = self._im2vec_encoder.get_output_dim()

        if dropout:
            self._dropout = torch.nn.Dropout(dropout)
        else:
            self._dropout = None
        if num_labels:
            self._num_labels = num_labels
        else:
            self._num_labels = vocab.get_vocab_size(namespace=label_namespace)
        self._classification_layer = torch.nn.Linear(self._classifier_input_dim, self._num_labels)
        self._accuracy = CategoricalAccuracy()
        self._loss = torch.nn.CrossEntropyLoss()
        initializer(self)

    def forward(self,  # type: ignore
                image: torch.FloatTensor,
                label: torch.IntTensor = None) -> Dict[str, torch.Tensor]:
        # pylint: disable=arguments-differ

        output = image
        if self._im2im_encoder:
            output = self._im2im_encoder(output)

        output = self._im2vec_encoder(output)

        if self._dropout:
            output = self._dropout(output)

        logits = self._classification_layer(output)
        probs = torch.nn.functional.softmax(logits, dim=-1)

        output_dict = {"logits": logits, "probs": probs}

        if label is not None:
            loss = self._loss(logits, label.long().view(-1))
            output_dict["loss"] = loss
            self._accuracy(logits, label)

        return output_dict

    @overrides
    def decode(self, output_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        output_dict['prediction'] = torch.argmax(output_dict['probs'], dim=1)
        all_predictions = output_dict['prediction']
        all_predictions = all_predictions.cpu().data.numpy()
        if self.vocab is not None:
            idx2token = self.vocab.get_index_to_token_vocabulary(namespace="labels")
            output_dict['class'] = [idx2token[x] for x in all_predictions]
        return output_dict

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = {'accuracy': self._accuracy.get_metric(reset)}
        return metrics



from allencv.common.testing import AllenCvTestCase, ModelTestCase
from allencv.data.dataset_readers import ImageClassificationDirectory
from allencv.models import BasicImageClassifier
from allencv.modules.im2im_encoders import FeedforwardEncoder
from allencv.modules.im2vec_encoders import FlattenEncoder


class TestBasicImageClassifier(ModelTestCase):

    def test_basic_experiment(self):
        data_directory = AllenCvTestCase.FIXTURES_ROOT / "data" / "image_classification"
        self.set_up_model(AllenCvTestCase.FIXTURES_ROOT / 'basic_image_classifier' / 'experiment.jsonnet',
                          data_directory)
        self.ensure_model_can_train_save_and_load(self.param_file)




from allennlp.common import Registrable


class ImageTransform(Registrable):

    def __init__(self, transform):
        self.transform = transform

    def __call__(self, **kwargs):
        self.transform(**kwargs)



from allencv.common.testing import AllenCvTestCase, ModelTestCase
from allencv.data.dataset_readers import PairedImageReader
from allencv.models import SemanticSegmentationModel
from allencv.modules.image_encoders import ResnetEncoder, FPN
from allencv.modules.image_decoders import BasicDecoder


class TestSemanticSegmentation(ModelTestCase):

    def test_basic_experiment(self):
        data_directory = AllenCvTestCase.FIXTURES_ROOT / "data" / "image_mask_reader"
        self.set_up_model(AllenCvTestCase.FIXTURES_ROOT / 'semantic_segmentation' / 'experiment.jsonnet',
                          data_directory)
        self.ensure_model_can_train_save_and_load(self.param_file)



from typing import Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F

from allencv.modules.image_encoders import ImageEncoder


@ImageEncoder.register("feature_pyramid")
class FPN(ImageEncoder):
    """
    A feature pyramid network.

    Given a list of feature maps at multiple spatial resolutions encoded  by a
    particular backbone encoder, FPN upscales each and merges it with higher level features.
    This produces a set of feature maps that have both coarse and fine-grained information
    with a large receptive field.

    The backbone must provide relative scale factors for its feature maps.

    Reference: https://arxiv.org/abs/1612.03144

    Parameters
    ----------
    backbone: ``ImageEncoder``
        Backbone that produces feature maps at various scales.
    output_channels: ``int``
        The number of output channels in each output feature map.
    """
    def __init__(self,
                 backbone: ImageEncoder,
                 output_channels: int):
        super(FPN, self).__init__()
        self._backbone = backbone
        self._upscale_factors = [a / b for a, b in zip(self._backbone.get_output_scales()[1:],
                                                       self._backbone.get_output_scales()[:-1])]
        self._output_channels = output_channels
        self.layers = nn.ModuleDict()
        self._convert_layers = nn.ModuleList()
        self._combine_layers = nn.ModuleList()

        for i, n_channels in enumerate(self._backbone.get_output_channels()):
            self._convert_layers.append(nn.Conv2d(n_channels, self._output_channels,
                                                       kernel_size=1, stride=1))
            self._combine_layers.append(
                nn.Conv2d(self._output_channels, self._output_channels, kernel_size=3, stride=1,
                          padding=1))

    def forward(self, image: torch.Tensor) -> Sequence[torch.Tensor]:
        outputs = self._backbone.forward(image)

        n = len(outputs)
        converted_images = [self._convert_layers[-1].forward(outputs[-1])]
        out = converted_images[-1]
        for i in range(n - 2, -1, -1):
            converted = self._convert_layers[i].forward(outputs[i])
            upscaled = F.interpolate(out, size=(converted.shape[-2], converted.shape[-1]))
            out = converted + upscaled
            converted_images.append(out)
        combined_images = [layer.forward(im) for layer, im in
                           zip(reversed(self._combine_layers), converted_images)]
        combined_images.insert(0,
                               F.max_pool2d(combined_images[0], kernel_size=1, stride=2, padding=0))

        return list(reversed(combined_images))

    def _get_test_outputs(self, height: int = 128, width: int = 128) -> Sequence[torch.Tensor]:
        params = list(self.parameters())
        im = torch.randn(1, self.get_input_channels(), height, width).to(params[0].device)
        return self.forward(im)

    def get_input_channels(self) -> int:
        return self._backbone.get_input_channels()

    def get_output_channels(self) -> Sequence[int]:
        return [x.shape[1] for x in self._get_test_outputs()]

    def get_output_scales(self) -> Sequence[int]:
        return self._backbone.get_output_scales()




from overrides import overrides

import torch
import torch.nn as nn
from torchvision.models.resnet import ResNet, resnet34, resnet50, resnet18, resnet101, resnet152

from allencv.modules.im2im_encoders import Im2ImEncoder

# class FromImageEncoder(Im2ImEncoder):
#     """
#     An ``Im2ImEncoder`` that passes an input image through a Resnet model structure.
#
#     Parameters
#     ----------
#     resnet_model: ``ResNet``
#         The base Resnet model.
#     layers: ``int``
#         The input will be passed through only the first ``layers`` layers of the
#         base Resnet model.
#     """
#     def __init__(self,
#                  image_encoder: ImageEncoder,
#                  index: int = -1) -> None:
#         self._encoder = image_encoder
#         self._index = index
#
#     def forward(self, inputs: torch.Tensor) -> torch.Tensor:
#         # pylint: disable=arguments-differ
#         output = self._encoder(inputs)[self._index]
#         return output
#
#     @overrides
#     def get_input_channels(self) -> int:
#         return self._encoder.get_input_channels()
#
#     @overrides
#     def get_output_channels(self) -> int:
#         return self._encoder.get_output_channels()[self._index]


class ResnetEncoder(Im2ImEncoder):
    """
    An ``Im2ImEncoder`` that passes an input image through a Resnet model structure.

    Parameters
    ----------
    resnet_model: ``ResNet``
        The base Resnet model.
    layers: ``int``
        The input will be passed through only the first ``layers`` layers of the
        base Resnet model.
    """
    def __init__(self,
                 resnet_model: ResNet,
                 layers: int = 4) -> None:
        super(ResnetEncoder, self).__init__()
        self._input_channels = 3
        res_layers = [resnet_model.layer1, resnet_model.layer2,
                      resnet_model.layer3, resnet_model.layer4]
        self._output_channels = res_layers[layers - 1][0].conv1.out_channels
        self.model = nn.Sequential(*([resnet_model.conv1,
                                    resnet_model.bn1,
                                    resnet_model.relu,
                                    resnet_model.maxpool] + res_layers[:layers] +
                                     [resnet_model.avgpool]))

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        # pylint: disable=arguments-differ
        output = self.model(inputs)
        return output

    @overrides
    def get_input_channels(self) -> int:
        return self._input_channels

    @overrides
    def get_output_channels(self) -> int:
        return self._output_channels


@Im2ImEncoder.register("pretrained_resnet")
class PretrainedResnetEncoder(ResnetEncoder):
    """
    Parameters
    ----------
    resnet_model: ``str``
        Name of the pretrained Resnet variant.
    requires_grad: ``bool``, optional (default = ``False``)
        Whether to continue training the Resnet model.
    num_layers: ``int``, optional (default = ``4``)
        How many of the 4 Resnet layers to include in the encoder.
    """
    def __init__(self, resnet_model: str, requires_grad: bool = False, num_layers: int = 4):
        if resnet_model == 'resnet34':
            model = resnet34(pretrained=True)
        elif resnet_model == 'resnet18':
            model = resnet18(pretrained=True)
        elif resnet_model == 'resnet152':
            model = resnet152(pretrained=True)
        elif resnet_model == 'resnet50':
            model = resnet50(pretrained=True)
        elif resnet_model == 'resnet101':
            model = resnet101(pretrained=True)
        else:
            raise ValueError(f"Model {resnet_model} is not supported.")

        for param in model.parameters():
            param.requires_grad = requires_grad

        super(PretrainedResnetEncoder, self).__init__(model, num_layers)



# pylint: disable=no-self-use,invalid-name
from allencv.common.testing.test_case import AllenCvTestCase
from allencv.data.dataset_readers import ImageAnnotationReader
from allencv.data.fields import ImageField, BoundingBoxField

from allennlp.data.fields import ListField


class TestImageAnnotationReader(AllenCvTestCase):

    def test_reader_output(self):
        reader = ImageAnnotationReader()
        instances = reader.read(self.FIXTURES_ROOT / "data" / "image_annotation")
        assert(isinstance(instances, list))
        assert len(instances) == 1
        fields = instances[0].fields
        assert isinstance(fields['image'], ImageField)
        assert isinstance(fields['boxes'], ListField)
        assert isinstance(fields['boxes'].field_list[0], BoundingBoxField)

    def test_missing_annotations(self):
        reader = ImageAnnotationReader(augmentation=[], annotation_dir='xyz')
        instances = reader.read(self.FIXTURES_ROOT / "data" / "image_annotation")
        assert(isinstance(instances, list))
        assert len(instances) == 1
        fields = instances[0].fields
        assert isinstance(fields['image'], ImageField)





from typing import Sequence, Union

import torch
from torchvision.models.resnet import resnet34, resnet50, resnet18, resnet101, resnet152, ResNet

from allencv.modules.image_encoders import ImageEncoder
import torch.nn as nn


@ImageEncoder.register("resnet_encoder")
class ResnetEncoder(ImageEncoder):
    """
        An ``ImageEncoder`` that passes an input image through a Resnet model structure.

        Parameters
        ----------
        resnet_model: ``ResNet``
            The base Resnet model.
        pretrained: ``bool``
            Use pretrained model from torchvision.
        requires_grad: ``bool``
            Shuts off backprop for this module.
        """

    def __init__(self,
                 resnet_model: Union[ResNet, str],
                 pretrained: bool = False,
                 requires_grad: bool = True,
                 norm_layer: nn.Module = None,
                 last_layer_maxpool: bool = False) -> None:
        super(ResnetEncoder, self).__init__()
        if isinstance(resnet_model, str):
            resnet_model = ResnetEncoder._pretrained_from_string(resnet_model,
                                                                 pretrained,
                                                                 requires_grad,
                                                                 norm_layer)
        self._input_channels = 3
        self.conv1 = resnet_model.conv1
        self.bn1 = resnet_model.bn1
        self.relu = resnet_model.relu
        self.maxpool = resnet_model.maxpool
        self.layer1 = resnet_model.layer1
        self.layer2 = resnet_model.layer2
        self.layer3 = resnet_model.layer3
        self.layer4 = resnet_model.layer4
        self._layers = [self.layer1, self.layer2, self.layer3, self.layer4]

    def forward(self, image: torch.Tensor) -> Sequence[torch.Tensor]:
        out = image
        out = self.maxpool(self.relu(self.bn1(self.conv1(out))))
        out_images = []
        for stage in self._layers:
            out = stage.forward(out)
            out_images.append(out)
        # TODO: not returning the first image here?
        return out_images

    def _get_test_outputs(self, height: int = 128, width: int = 128) -> Sequence[torch.Tensor]:
        params = list(self.parameters())
        im = torch.randn(1, self.get_input_channels(), height, width).to(params[0].device)
        return self.forward(im)

    def get_output_channels(self) -> Sequence[int]:
        return [x.shape[1] for x in self._get_test_outputs()]

    def get_input_channels(self) -> int:
        return 3

    def get_output_scales(self) -> Sequence[int]:
        test_height = 128
        test_width = 128
        outputs = self._get_test_outputs(test_height, test_width)
        return [test_height // out.shape[-2] for out in outputs]

    @staticmethod
    def _pretrained_from_string(resnet_string: str,
                                pretrained: bool = False,
                                requires_grad: bool = False,
                                norm_layer: nn.Module = None):
        if resnet_string == 'resnet34':
            model = resnet34(pretrained=pretrained, norm_layer=norm_layer)
        elif resnet_string == 'resnet18':
            model = resnet18(pretrained=pretrained, norm_layer=norm_layer)
        elif resnet_string == 'resnet152':
            model = resnet152(pretrained=pretrained, norm_layer=norm_layer)
        elif resnet_string == 'resnet50':
            model = resnet50(pretrained=pretrained, norm_layer=norm_layer)
        elif resnet_string == 'resnet101':
            model = resnet101(pretrained=pretrained, norm_layer=norm_layer)
        else:
            raise ValueError(f"Model {resnet_string} is not supported.")
        for param in model.parameters():
            param.requires_grad = requires_grad
        return model




from setuptools import find_packages, setup

setup(
    name='allencv',
    packages=find_packages(),
    version='0.1.2',
    description='',
    author='sethah',
    license='MIT',
    python_requires='>=3.6.1',
    install_requires=[
        'torch>=0.4.1',
        'torchvision==0.3.0',
        'allennlp',
        'numpy',
        'overrides',
        'ipython',
        'albumentations',
        'Pillow',
        'overrides',
        'numpy',
      ],
    tests_require=[
        'pytest'
    ]
)



import numpy as np
from overrides import overrides
from pathlib import Path
from PIL import Image

from typing import Dict, Iterable, List
import logging

import torch

from allencv.data.transforms.image_transform import ImageTransform
from allencv.data.fields.image_field import ImageField
from allencv.data.dataset_readers.image_dataset_reader import ImageDatasetReader

from allennlp.data.instance import Instance
from allennlp.data.fields import LabelField, Field
from allennlp.data.dataset_readers import DatasetReader

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@DatasetReader.register("image_classification_directory")
class ImageClassificationDirectory(ImageDatasetReader):
    """
    Reads images from directories where the directory name is the label name for
    all images contained within. Only works one level deep currently.

    data_path
    |-- Cat
    |   |-- 00001.jpg
    |   |-- 00002.jpg
    |   |-- ...
    |-- Dog
    |   |-- 00001.json
    |   |-- 00002.json
    |   |-- ...

    Parameters
    ----------
    skip_label_indexing: ``bool``, optional (default = ``False``)
        Whether or not to skip label indexing. You might want to skip label indexing if your
        labels are numbers, so the dataset reader doesn't re-number them starting from 0.
    lazy : ``bool``, optional, (default = ``False``)
        Whether or not instances can be read lazily.
    """
    def __init__(self,
                 augmentation: List[ImageTransform] = None,
                 skip_label_indexing: bool = False,
                 lazy: bool = False) -> None:
        super(ImageClassificationDirectory, self).__init__(augmentation)
        self.lazy = lazy
        self._skip_label_indexing = skip_label_indexing

    def _read(self, file_path: str) -> Iterable[Instance]:
        for label_dir in Path(file_path).iterdir():
            label = label_dir.name
            for img_file in label_dir.iterdir():
                if img_file.suffix not in ['.jpg', '.png']:
                    continue
                with open(img_file, 'rb') as f:
                    try:
                        img = Image.open(f)
                    except (UnboundLocalError, OSError) as e:
                        logger.error(f"Couldn't read {img_file}")
                        continue
                    sample = img.convert('RGB')
                    yield self.text_to_instance(np.array(sample), label)

    @overrides
    def text_to_instance(self, image: np.ndarray, label: str = None) -> Instance:
        image, _, _ = self.augment(image)
        fields: Dict[str, Field] = {}
        fields['image'] = ImageField(image.transpose(2, 0, 1), channels_first=False)
        if label is not None:
            fields['label'] = LabelField(label, skip_indexing=self._skip_label_indexing)
        return Instance(fields)



from allencv.modules.im2im_encoders.im2im_encoder import Im2ImEncoder
from allencv.modules.im2im_encoders.feedforward_encoder import FeedforwardEncoder
from allencv.modules.im2im_encoders.resnet_encoder import ResnetEncoder
